(***************************************************************************
  EditList unit
  Edit a list of items, user definable
  PJB January 9, 1994, Internet mail to d91-pbr@nada.kth.se
  Copyright PJB 1994, All Rights Reserved.
  Free source, use at your own risk.
  If modified, please state so if you pass this around.

  This unit provides the basic list handling for displaying a list
  of items that can be edited, copied, deleted and added.

  You need a list item manager that takes care the actual editing
  and possibly some other things.

   The list edit dialog needs at least one element to create new
    elements, so you must always have an empty topic at the end of the
    list if you have a New button.

  Two examples are provided, a Tool dialog and a Window list, both
  like BP's. To see these, run TOOLTEST.

  The code is admittedly tuned to work with the above examples and
  might need more work to be truly reusable in another context.

***************************************************************************)
unit EditList;
{$B-,X+}

interface

  uses
    App, Dialogs, Drivers, Objects, Views,
    toyPrefs, {$I hcFile}
    toyUtils;

  type
    FKeySet     = set of 1..10;
    LEOptionSet = set of (leLastItemEmpty, leDelKey, leNew, leEdit);

    { Used by TListItemBox }
    PGetText = ^TGetText;
    TGetText =
      object (TObject)
        function GetText:String; virtual;
      end;

    { These items are what EditListDialog manages }
    PListItem = ^TListItem;
    TListItem =
      object (TGetText)
        procedure Delete; virtual;
        function  Edit(New:Boolean):PListItem; virtual;
      end;

    { ListBox with overridden GetText }
    PListItemBox = ^TListItemBox;
    TListItemBox =
      object (TListBox)
        function GetText(Item:Integer; MaxLen:Integer):String; virtual;
      end;

    { This is the main list editing dialog }
    PEditListDialog = ^TEditListDialog;
    TEditListDialog =
      object (TDialog)
        ListBox   : PListItemBox;
        MaxCount  : Integer;
        LEOptions : LEOptionSet;

        constructor Init(aTitle, aLabel:String;
            Width, Height, aMaxCount:Integer; aLEOptions:LEOptionSet);
        constructor InitHC(aTitle, aLabel:String;
            Width, Height, aMaxCount:Integer; aLEOptions:LEOptionSet; HelpContext:Word);
        constructor Load(var S: TStream);
        procedure HandleEvent(var Event: TEvent); virtual;
        procedure Store(var S: TStream);
      end;


    { Tool editing manager }
    PToolListItem = ^TToolListItem;
    TToolListItem =
      object (TListItem)
        Title       : String[32];
        Path        : String[128];
        CommandLine : String;
        HotKey      : Word;

        constructor InitData(const aTitle, aPath, aCommandLine:String; aHotkey:Word);
        constructor Load(var S:TStream);
        procedure Delete; virtual;
        function  Dlg(DialogTitle:String):Word;
        function  Edit(New:Boolean):PListItem; virtual;
        function  GetText:String; virtual;
        procedure Store(var S:TStream);
      end;

    { Window list manager }
    PWindowListItem = ^TWindowListItem;
    TWindowListItem =
      object (TListItem)
        Window:PWindow;
        constructor Init(AWindow:PWindow);
        procedure Delete; virtual;
        function  GetText:String; virtual;
      end;

  var
    (* Function keys currently in use by tools *)
    toolUsedFKeys : FKeySet;

  const
    (* Function keys possible to use for tools: F1..F10, except F6 *)
    toolAllowedFKeys : FKeySet = [1..10]-[6];
    (* The name of the shift key for tool hot key access *)
    toolShiftKey : String[5] = 'Shift';

    (* The standard tool dialog's options *)
    leStandardToolDialog   = [leLastItemEmpty, leNew, leEdit, leDelKey];
    (* The standard window list dialog's options *)
    leStandardWindowDialog = [leDelKey];


  const
    RListItemBox : TStreamRec = (
      ObjType : 12345;            {<--- Insert a unique number >= 100 here!!}
      VmtLink : Ofs(Typeof(TListItemBox)^);
      Load    : @TListItemBox.Load;
      Store   : @TListItemBox.Store);

    REditListDialog : TStreamRec = (
      ObjType : 12345;            {<--- Insert a unique number >= 100 here!!}
      VmtLink : Ofs(Typeof(TEditListDialog)^);
      Load    : @TEditListDialog.Load;
      Store   : @TEditListDialog.Store);

  const
    RToolListItem : TStreamRec = (
      ObjType : 12345;            {<--- Insert a unique number >= 100 here!!}
      VmtLink : Ofs(Typeof(TToolListItem)^);
      Load    : @TToolListItem.Load;
      Store   : @TToolListItem.Store);


  function CalcHotKey(HotKey:Word):Word;
  function HotKeyStr(HotKey:Word):String;
  function kbHotKey(BaseKey, HotKey:Word):Word;

  function MakeStandardToolDialog(MaxTools:Integer):PDialog;
  function MakeStandardWindowListDialog:PDialog;

  function MakeToolEditDialog(const Title, ShiftKey:String; Keys, UsedKeys:FKeySet):PDialog;

(***************************************************************************
***************************************************************************)
implementation


  const
    { Modal commands }
    cmtoyNewItem               = 255;
    cmtoyEditItem              = 1000;
    cmtoyDeleteItem            = 1002;

  const
    (* Relative list editing help contexts *)
    hctoyLEList                = 0;
    hctoyLEEdit                = 1;
    hctoyLENew                 = 2;
    hctoyLEDelete              = 3;
    hctoyLEListEditDialogHelp  = 4;


  (*******************************************************************
    List box treats items as TGetText and uses their GetText
  *******************************************************************)
  function TListItemBox.GetText(Item: Integer; MaxLen: Integer): String;
  begin
    GetText:=PGetText(List^.At(Item))^.GetText;
  end;


  (*******************************************************************
    How not to write a general GetText
    This one assumes the first field in the TObject is a string
  *******************************************************************)
  function TGetText.GetText:String;
  begin
    GetText:=PString(PChar(@Self)+2)^;
  end;


  (*******************************************************************
  *******************************************************************)
  procedure TListItem.Delete; begin end;
  function TListItem.Edit; begin end;


(***************************************************************************
***************************************************************************)

  (*******************************************************************
    Create list editing dialog, no specific help context
  *******************************************************************)
  constructor TEditListDialog.Init;
  begin
    InitHC(aTitle, aLabel, Width, Height, aMaxCount, aLEOptions, hcNoContext);
  end;


  (*******************************************************************
    Create list editing dialog
  *******************************************************************)
  constructor TEditListDialog.InitHC;
    var
      R: TRect;
      Control : PView;
  begin
    R.Assign(18, 3, 37+Width, 8+Height);
    inherited Init(R, ATitle);
    Options := Options or ofCentered;
    MaxCount:=aMaxCount;
    LEOptions := aLEOptions;

    R.Assign(3+Width, 3, 4+Width, 3+Height);
    Control := New(PScrollBar, Init(R));
    Insert(Control);

    R.Assign(3, 3, 3+Width, 3+Height);
    New(ListBox, Init(R, 1, PScrollbar(Control)));
    ListBox^.HelpCtx := HelpContext+hctoyLEList;
    Insert(ListBox);

    R.Assign(2, 2, 17, 3);
    Insert(New(PLabel, Init(R, ALabel, ListBox)));

    R.Assign(6+Width, 3, 16+Width, 5);
    Control := New(PButton, Init(R, 'O~K~', cmOK, bfDefault-Ord(leEdit in LEOptions)));
    Control^.HelpCtx := hcOK;
    Insert(Control);

    if leEdit in LEOptions then
    begin
      R.Move(0, 2);
      Control := New(PButton, Init(R, '~E~dit', cmtoyEditItem, bfDefault));
      Control^.HelpCtx := HelpContext+hctoyLEEdit;
      Insert(Control);
    end;

    if leNew in LEOptions then
    begin
      R.Move(0, 2);
      Control := New(PButton, Init(R, '~N~ew', cmtoyNewItem, bfNormal));
      Control^.HelpCtx := HelpContext+hctoyLENew;
      Insert(Control);
    end;

    R.Move(0, 2);
    Control := New(PButton, Init(R, '~D~elete', cmtoyDeleteItem, bfNormal));
    Control^.HelpCtx := HelpContext+hctoyLEDelete;
    Insert(Control);

    R.Move(0, 2);
    Control := New(PButton, Init(R, 'Cancel', cmCancel, bfNormal));
    Control^.HelpCtx := hcCancel;
    Insert(Control);

    R.Move(0, 2);
    Control := New(PButton, Init(R, 'Help', cmHelp, bfGrabFocus));
    Control^.HelpCtx := HelpContext+hctoyLEListEditDialogHelp;
    Insert(Control);

    SelectNext(False);
  end;


  (*******************************************************************
    Load from stream
  *******************************************************************)
  constructor TEditListDialog.Load(var S:TStream);
  begin
    inherited Load(S);
    GetSubViewPtr(S, ListBox);
    S.Read(LEOptions, SizeOf(LEOptions));
    S.Read(MaxCount, SizeOf(MaxCount));
  end;


  (*******************************************************************
    Handle events
  *******************************************************************)
  procedure TEditListDialog.HandleEvent(var Event:TEvent);
    var
      Index    : Integer;
      Count    : Integer;
      Item     : PListItem;

    procedure DeleteItem;
    begin
      if Index<Count-Ord(leLastItemEmpty in LEOptions) then
      begin
        Item^.Delete;
        ListBox^.List^.AtFree(Index);
        ListBox^.SetRange(Count-1);
      end;
    end;

    procedure EditItem(New:Boolean);
      var
        P : PListItem;
    begin
      if not ((leLastItemEmpty in LEOptions) and (Count>MaxCount) and
        (Index=Count-1)) and (Index<Count) then
      begin
        P:=Item^.Edit(New);
        if P<>Nil then
          with ListBox^ do
          begin
            List^.AtInsert(Count-Ord(leLastItemEmpty in LEOptions), P);
            SetRange(Count+1);
            FocusItem(Count-Ord(leLastItemEmpty in LEOptions));
          end;
      end;
    end;

  begin
    inherited HandleEvent(Event);

    Index:=ListBox^.Focused;
    Count:=ListBox^.List^.Count;
    if Count>0 then
      Item:=ListBox^.List^.At(Index);

    if Count>=MaxCount+Ord(leLastItemEmpty in LEOptions) then
    begin
      DisableCommands([cmtoyNewItem]);
      CommandSetChanged:=True;
    end
    else
    begin
      EnableCommands([cmtoyNewItem]);
      CommandSetChanged:=True;
    end;

    case Event.What of
      evBroadcast,
      evCommand:
        case Event.Command of
          cmtoyDeleteItem:      DeleteItem;
          cmListItemSelected,
          cmtoyEditItem:        EditItem(False);
          cmtoyNewItem:         EditItem(True);
          else
            Exit;
        end;
      evKeyboard:
        case Event.KeyCode of
          kbDel:           if leDelKey in LEOptions then DeleteItem;
          else
            Exit;
        end;
      else
        Exit;
    end;
    ListBox^.Draw;
    ClearEvent(Event);
  end;


  (*******************************************************************
    Store on stream
  *******************************************************************)
  procedure TEditListDialog.Store(var S:TStream);
  begin
    inherited Store(S);
    PutSubViewPtr(S, ListBox);
    S.Write(LEOptions, SizeOf(LEOptions));
    S.Write(MaxCount, SizeOf(MaxCount));
  end;


(***************************************************************************
***************************************************************************)

  (*******************************************************************
    Edit a tool dialog
  *******************************************************************)
  function MakeToolEditDialog;
    var
      FCount : Integer;

    (*******************************************************************
      Create the hot key radio buttons
    *******************************************************************)
    function CreateFunctionKeys:PSItem;
      var
        i : Integer;
        P : PSItem;
    begin
      FCount:=0;

      if 10 in Keys then
      begin
        P:=NewSItem(ShiftKey+'+F1~0~', Nil);
        Inc(FCount);
      end
      else
        P:=Nil;

      for i:=9 downto 1 do
        if i in Keys then
        begin
          P:=NewSItem(ShiftKey+'+F~'+Chr(Ord('0')+i)+'~', P);
          Inc(FCount);
        end;

      CreateFunctionKeys:=P;
    end;

    var
      Dlg : PDialog;
      R : TRect;
      Control : PView;
      DY : Integer;
  begin
    R.Assign(12, 6, 68, 24);
    New(Dlg, Init(R, Title));
    Dlg^.Options := Dlg^.Options or ofCentered;

    R.Assign(3, 3, 35, 4);
    Control := New(PInputLine, Init(R, 32));
    Control^.HelpCtx := hctoyTEProgram;
    Dlg^.Insert(Control);

    R.Assign(2, 2, 8, 3);
    Dlg^.Insert(New(PLabel, Init(R, '~T~itle', Control)));

    R.Assign(3, 6, 35, 7);
    Control := New(PInputLine, Init(R, 128));
    Control^.HelpCtx := hctoyTEPath;
    Dlg^.Insert(Control);

    R.Assign(2, 5, 15, 6);
    Dlg^.Insert(New(PLabel, Init(R, 'Program ~p~ath', Control)));

    R.Assign(3, 9, 35, 10);
    Control := New(PInputLine, Init(R, 255));
    Control^.HelpCtx := hctoyTECommandLine;
    Dlg^.Insert(Control);

    R.Assign(2, 8, 15, 9);
    Dlg^.Insert(New(PLabel, Init(R, 'Command ~l~ine', Control)));

    R.Assign(37, 3, 53, 4);
    Control := New(PRadioButtons, Init(R,
      NewSItem('~U~nassigned',
      CreateFunctionKeys)));
    Control^.HelpCtx := hctoyTEHotKey;

    (* I'm surprised this works, seems Word([1]) is actually 2! *)
    Dec(LongRec(PRadioButtons(Control)^.EnableMask).Lo, Word(UsedKeys));
    Inc(R.B.Y, FCount);
    Control^.ChangeBounds(R);
    Dlg^.Insert(Control);

    R.Assign(36, 2, 44, 3);
    Dlg^.Insert(New(PLabel, Init(R, '~H~ot key', Control)));

    DY:=Min(4, 10-FCount);
    Dlg^.GrowTo(Dlg^.Size.X, Dlg^.Size.Y-DY);

    R.Assign(20, 15-DY, 30, 17-DY);
    Control := New(PButton, Init(R, 'O~K~', cmOK, bfDefault));
    Control^.HelpCtx := hcOK;
    Dlg^.Insert(Control);

    R.Move(12, 0);
    Control := New(PButton, Init(R, 'Cancel', cmCancel, bfNormal));
    Control^.HelpCtx := hcCancel;
    Dlg^.Insert(Control);

    R.Move(12, 0);
    Control := New(PButton, Init(R, 'Help', cmHelp, bfGrabFocus));
    Control^.HelpCtx := hctoyTEDialogHelp;
    Dlg^.Insert(Control);

    Dlg^.SelectNext(False);
    MakeToolEditDialog := Dlg;
  end;


(***************************************************************************
***************************************************************************)

  (*******************************************************************
    Create a tool
  *******************************************************************)
  constructor TToolListItem.InitData(const aTitle, aPath, aCommandLine:String; aHotkey:Word);
  begin
    Title:=aTitle;
    Path:=aPath;
    CommandLine:=aCommandLine;
    HotKey:=aHotKey;
  end;


  (*******************************************************************
    Load from stream
  *******************************************************************)
  constructor TToolListItem.Load(var S:TStream);
  begin
    S.Read(Title, SizeOf(Title)+SizeOf(Path)+SizeOf(CommandLine)+SizeOf(HotKey));
    if HotKey<>0 then
      Include(toolUsedFKeys, HotKey);
  end;


  (*******************************************************************
    Delete tool
  *******************************************************************)
  procedure TToolListItem.Delete;
  begin
    if HotKey<>0 then
      Exclude(toolUsedFKeys, HotKey);
  end;


  (*******************************************************************
    Edit tool dialog
  *******************************************************************)
  function TToolListItem.Dlg(DialogTitle:String):Word;
    var
      cm : Word;
  begin
    cm:=Application^.ExecuteDialog(MakeToolEditDialog(
        DialogTitle, toolShiftKey, toolAllowedFKeys, toolUsedFKeys-[HotKey]), @Title);

    if (cm=cmOK) and (HotKey<>0) then
      Include(toolUsedFKeys, HotKey);

    Dlg:=cm;
  end;


  (*******************************************************************
    Edit or create a tool
    Tools with empty titles are not saved, same as in BP
  *******************************************************************)
  function TToolListItem.Edit;
    var
      P : PToolListItem;
  begin
    Edit:=Nil;

    if New or (Title='') then            (* Editing new tool *)
    begin
      System.New(P, Init);
      P^:=Self;
      P^.HotKey:=0;

      if (P^.Dlg('New Tool')=cmOK) and (P^.Title<>'') then
        Edit:=P
      else
        Dispose(P, Done);
    end
    else
      Dlg('Modify Tool');
  end;


  (*******************************************************************
    Return text to show in selection list box
    This one just strips out tildes (~)
  *******************************************************************)
  function TToolListItem.GetText:String;
    var
      i,j : Integer;
  begin
    j:=0;
    for i:=1 to Length(Title) do
      if Title[i]<>'~' then
      begin
        Inc(j);
        GetText[j]:=Title[i]
      end;
    GetText[0]:=Chr(j);
  end;


  (*******************************************************************
    Store on stream
  *******************************************************************)
  procedure TToolListItem.Store(var S:TStream);
  begin
    S.Write(Title, SizeOf(Title)+SizeOf(Path)+SizeOf(CommandLine)+SizeOf(HotKey));
  end;


(***************************************************************************
***************************************************************************)

   (*******************************************************************
     Store window pointer
   *******************************************************************)
   constructor TWindowListItem.Init(AWindow:PWindow);
   begin
     Window:=AWindow;
   end;


   (*******************************************************************
     Close the window
   *******************************************************************)
   procedure TWindowListItem.Delete;
   begin
     Message(Window, evCommand, cmClose, Nil);
   end;


   (*******************************************************************
     Return window title
   *******************************************************************)
   function TWindowListItem.GetText:String;
   begin
     GetText:=Window^.Title^;
   end;


(***************************************************************************
***************************************************************************)

  (*******************************************************************
    Calulate true hot key from check box value
  *******************************************************************)
  function CalcHotKey(HotKey:Word):Word;
    var
      i : Word;
  begin
    for i:=1 to 10 do
      if (i<=HotKey) and not (i in toolAllowedFKeys) then
        Inc(HotKey);
    CalcHotKey:=HotKey;
  end;


  (*******************************************************************
    Return a hot key string based on ordinal hotkey
  *******************************************************************)
  function HotKeyStr(HotKey:Word):String;
  begin
    HotKey:=CalcHotKey(HotKey);
    if HotKey=0 then
      HotKeyStr:=''
    else
      HotKeyStr:=toolShiftKey+'+F'+ToStr(HotKey);
  end;


  (*******************************************************************
    Return a hot key's kbXXX value
  *******************************************************************)
  function kbHotKey(BaseKey, HotKey:Word):Word;
  begin
    HotKey:=CalcHotKey(HotKey);
    if HotKey=0 then
      kbHotKey:=kbNoKey
    else
      kbHotKey:=BaseKey+(HotKey-1)*256;
  end;


  (*******************************************************************
    Create standard tools dialog
  *******************************************************************)
  function MakeStandardToolDialog(MaxTools:Integer):PDialog;
  begin
    MakeStandardToolDialog:=New(PEditListDialog,
      InitHC('Tools', '~P~rogram titles', 24, 11, MaxTools, leStandardToolDialog, hctoyToolEditList));
  end;


  (*******************************************************************
    Create standard window list dialog
  *******************************************************************)
  function MakeStandardWindowListDialog:PDialog;
  begin
    MakeStandardWindowListDialog:=New(PEditListDialog,
      InitHC('Window List', '~W~indows', 40, 10, MaxInt, leStandardWindowDialog, hctoyWindowList));
  end;


end.